// gradAccounts.cc
// version 1.0, SM, May 2021
//              based on BuHa by Lg
// version 1.1, SM, May 2021
// version 1.2, SM, May 2021
// version 1.3, SM, May 2021
// version 1.4, SM, May 2021
// version 1.5, SM, June 2021
// version 1.7, SM, June 2021
// version 1.8, SM, July 2021
// version 1.9, SM, July 2021
// version 1.10, SM, August 2021
// version 1.12, SM, July 2022
// version 1.13, SM, December 2023

#include <iostream>
#include <fstream>
#include <iomanip>
#include <string.h>
#include <math.h>
#include "uti1a.h"
#include "grad.h"
using namespace std;

static char* empty_string = (char*)"";

static string latex_header1 = "\\documentclass[a4paper,12pt]{article}\n";
static string latex_header1l = "\\documentclass[a4paper,12pt,landscape]{article}\n";
static string latex_header2 = "\n\\usepackage[sf]{research18}\n\\pagestyle{empty}\n\\setfixedmargins{2cm}\n\n%% Deactivate underscore\n\\catcode`\\_=11\n\n\\begin{document}\n\\frenchspacing\n\n";
static string latex_footer = "\\end{document}\n\n%%% Local Variables:\n%%% mode: latex\n%%% TeX-master: t\n%%% End:\n\n";


void pointslist::put_entry( float points ) 
//               ========= 
{
  points_entry* pe = new points_entry;
  if ( pe==0 ) memory_error();
  append(pe);
  pe->_points = points;
  totalpoints += points;
}


void pointslist::clear()
{ //             =====
  points_entry *p = (points_entry*)_first;
  while (p) {
    points_entry *next = (points_entry*)p->_next;
    delete p;
    p = next;
  }
  _list_length = 0;
  _first = _last = 0;
}


void computegradelist::put_entry( exam* ep ) 
//                     =========             
{
  computegrade_entry* cgep = new computegrade_entry;
  if ( cgep==0 ) memory_error();
  append(cgep);
  cgep->_gradep = ep;
}


void computegradelist::clear()
{ //                   =====
  computegrade_entry *p = (computegrade_entry*)_first;
  while (p) {
    computegrade_entry *next = (computegrade_entry*)p->_next;
    delete p;
    p = next;
  }
  _list_length = 0;
  _first = _last = 0;
}



exam::exam( char* name, char* comment, int isexam, float weight,
/*    ==== */	float x1, float y1, float x2, float y2)
{
  _weight = weight; 
  strngp = name; 
  _x1 = x1;
  _x2 = x2;
  _y1 = y1;
  _y2 = y2;
  _isexam = isexam;
  _hiddenflag = 0;
  _numberofproblems=0;
  _exactsum=0;
  _exactsquaredsum=0;
  _sum=0;
  _squaredsum=0;
  _pointssump = new pointslist();
  if ( _pointssump==0 ) memory_error();
  _cglp = 0;
  if ( comment==0 ) {
    _comment = empty_string;
  } else {
    _comment = new char[strlen(comment)+1];
    if ( _comment==0 ) memory_error();
    strcpy( _comment, comment );
  }
}



void exam::put_gradeentry( student* studp, float grade ) 
/*         ============== */        
{
  exam_entry* p = new exam_entry;
  if ( p==0 ) memory_error();
  append(p);
  p->_studp = studp; 
  p->_examp = this; 
  p->_exactgrade = grade;
  p->_grade = grade;
  _exactsum += grade;
  _exactsquaredsum += grade*grade;
  _sum += grade;
  _squaredsum += grade*grade;
  student_entry* sep = new student_entry;
  if ( sep==0 ) memory_error();
  sep->eep = p;
  studp->append(sep);
}



void exam::overwrite_gradeentry( exam_entry* eep, float grade ) 
/*         ==================== */        
{
  float origgrade=eep->_grade;
  eep->_grade = grade;
  _sum = _sum - origgrade + grade;
  _squaredsum = _squaredsum - origgrade*origgrade + grade*grade;
}



int exam::put_examentry( student* studp, pointslist* pointsp, float rounding,
/*        ============= */     float maxgrade, float mingrade ) 
{
  int status = 0;
  exam_entry* p = new exam_entry;
  if ( p==0 ) memory_error();
  append(p);
  p->_studp = studp; 
  p->_examp = this; 
  p->_pointsp = pointsp;
  if ( _numberofproblems == 0 ) {
    _numberofproblems = pointsp->number_of_problems();
  } else if ( _numberofproblems != pointsp->number_of_problems() ) {
    if ( _numberofproblems < pointsp->number_of_problems() ) {
      _numberofproblems = pointsp->number_of_problems();
    }
    status = 1;
  }
  // add points to pointssum for pointsaverage
  simpleListEntry* pep = pointsp->_first;
  simpleListEntry* newpep = _pointssump->_first;
  while (pep) {
    if ( newpep == 0 ) {
      points_entry* pe = new points_entry;
      if ( pe==0 ) memory_error();
      _pointssump->append(pe);
      pe->_points = ((points_entry*)pep)->_points;
      _pointssump->totalpoints += ((points_entry*)pep)->_points;
    } else {
      ((points_entry*)newpep)->_points += ((points_entry*)pep)->_points;
      _pointssump->totalpoints += ((points_entry*)pep)->_points;
      newpep=newpep->_next;
    }
    pep=pep->_next;
  }
  p->_exactgrade = (_y2-_y1)/(_x2-_x1)*(pointsp->totalpoints-_x1) + _y1;
  if (rounding>0) {
    p->_grade = round((p->_exactgrade)/rounding)*rounding;
  } else {
    p->_grade = p->_exactgrade;
  }
  if (p->_grade > maxgrade) { p->_grade = maxgrade; }
  if (p->_grade < mingrade) { p->_grade = mingrade; }
  _exactsum += p->_exactgrade;
  _exactsquaredsum += p->_exactgrade * p->_exactgrade;
  _sum += p->_grade;
  _squaredsum += p->_grade * p->_grade;
  student_entry* sep = new student_entry;
  if ( sep==0 ) memory_error();
  sep->eep = p;
  studp->append(sep);
  return status;
}



exam_entry* exam::stud_in_exam_search(const char* studname)
{  //             ===================
  simpleListEntry* cp = _first;
  exam_entry* ep;
  student* curr_stud; 
  int i;
  char sti;
  while (cp) {
    ep = (exam_entry*)cp;
    curr_stud = ep->_studp;
    i=0;
    while ( (sti=studname[i]) == (curr_stud->strngp[i]) ) {
      if (sti) i++; 
      else {  return ep; }
    }
    cp = cp->_next;
  }
  return 0;
}



void exam::sort( int order )
{ //       ====
  simpleListEntry* oldcurrentp = _first;
  simpleListEntry* oldnextp;
  _first = 0;
  _last = 0;
  _list_length = 0;
  simpleListEntry* currentp;
  simpleListEntry* previousp;
  exam_entry* ep;
  exam_entry* oldep;
  while (oldcurrentp) {
    oldnextp = oldcurrentp->_next;
    oldep = (exam_entry*)oldcurrentp;
    currentp = _first;
    previousp = 0;
    while (currentp) {
      ep = (exam_entry*)currentp;
      if ( order == sortbygrade ) {
	if (_isexam) {
	  if ((ep->_pointsp->totalpoints < oldep->_pointsp->totalpoints)
	      || ((ep->_pointsp->totalpoints == oldep->_pointsp->totalpoints)
		  && (stringcomp(oldep->_studp->strngp,ep->_studp->strngp)))) {
	    break;
	  }
	} else {
	  if ((ep->_exactgrade < oldep->_exactgrade)
	      || ((ep->_exactgrade == oldep->_exactgrade)
		  && (stringcomp(oldep->_studp->strngp,ep->_studp->strngp)))) {
	    break;
	  }
	}
      } else if ( order == sortbyname ) {
	if (stringcomp(oldep->_studp->strngp,ep->_studp->strngp)) {
	  break;
	}
      } else if ( order == sortbycomment ) {
	if (stringcomp(oldep->_studp->index(),ep->_studp->index())) {
	  break;
	}
      } else {
	return;
      }
      previousp = currentp;
      currentp = currentp->_next;
    }
    insert(oldcurrentp,previousp);
    oldcurrentp = oldnextp;
  }
}



int exam::stringcomp( const char* first, const char* second )
{ //       =========
  // returns 1 if first <= second, 0 otherwise
  int i=0;
  char firsti, secondi;
  while ( (firsti=first[i]) == (secondi=second[i]) ) {
    if (firsti) i++; 
    else return 1; 
  }
  if ( firsti < secondi ) return 1;
  return 0;
}



void exam::add_avg( exam* from, float rounding, float maxgrade, float mingrade)
{ //       =======
  if (rounding == -1) {
    //we are in the process of adding grades to the total sum
    if (_cglp) {
      _cglp->put_entry(from);
    } else {
      _cglp = new computegradelist( computegradelist::isavg );
      _cglp->put_entry(from);
    }
    simpleListEntry* fp = from->_first;
    exam_entry* fep;
    exam_entry* tep;
    while (fp) {
      fep = (exam_entry*)fp;
      tep = stud_in_exam_search((fep->_studp)->strngp);
      if ( tep != 0 ) {
	tep->_exactgrade += fep->_grade * fep->_examp->weight();
	//temporary storage of total weight
	tep->_tmpweight += fep->_examp->weight(); 
      } else {
	exam_entry* eep = new exam_entry; 
	if ( eep==0 ) memory_error();
	append(eep);
	eep->_studp = fep->_studp;
	eep->_examp = this;
	eep->_exactgrade = fep->_grade * fep->_examp->weight();
	//temporary storage of total weight
	eep->_tmpweight = fep->_examp->weight();
	student_entry* sep = new student_entry;
	if ( sep==0 ) memory_error();
	sep->eep = eep;
	(fep->_studp)->append(sep);
      }
      fp = fp->_next;
    }
  } else {
    //adding has ended, now we need to compute the average grades
    simpleListEntry* tp = _first;
    exam_entry* tep;
    while (tp) {
      tep = (exam_entry*)tp;
      tep->_exactgrade = tep->_exactgrade/tep->_tmpweight;
      tep->_tmpweight = 0; //delete temporary total weight
      if (rounding>0) {
	tep->_grade = round((tep->_exactgrade)/rounding)*rounding;
      } else {
	tep->_grade = tep->_exactgrade;
      }
      if (tep->_grade > maxgrade) { tep->_grade = maxgrade; }
      if (tep->_grade < mingrade) { tep->_grade = mingrade; }
      _exactsum += tep->_exactgrade;
      _exactsquaredsum += tep->_exactgrade * tep->_exactgrade;
      _sum += tep->_grade;
      _squaredsum += tep->_grade * tep->_grade;
      tp = tp->_next;
    }
  }
}



void exam::add_max( exam* from, float rounding, float maxgrade, float mingrade)
{ //       =======
  if (rounding == -1) {
    //we are in the process of finding the maximum grade
    if (_cglp) {
      _cglp->put_entry(from);
    } else {
      _cglp = new computegradelist( computegradelist::ismax );
      _cglp->put_entry(from);
    }
    simpleListEntry* fp = from->_first;
    exam_entry* fep;
    exam_entry* tep;
    while (fp) {
      fep = (exam_entry*)fp;
      tep = stud_in_exam_search((fep->_studp)->strngp);
      if ( tep != 0 ) {
	if ( tep->_exactgrade < fep->_grade ) {
	  tep->_exactgrade = fep->_grade;
	}
      } else {
	exam_entry* eep = new exam_entry;
	if ( eep==0 ) memory_error();
	append(eep);
	eep->_studp = fep->_studp;
	eep->_examp = this;
	eep->_exactgrade = fep->_grade;
	student_entry* sep = new student_entry;
	if ( sep==0 ) memory_error();
	sep->eep = eep;
	(fep->_studp)->append(sep);
      }
      fp = fp->_next;
    }
  } else {
    //finding the maximum has ended, now we need to round if required
    simpleListEntry* tp = _first;
    exam_entry* tep;
    while (tp) {
      tep = (exam_entry*)tp;
      if (rounding>0) {
	tep->_grade = round((tep->_exactgrade)/rounding)*rounding;
      } else {
	tep->_grade = tep->_exactgrade;
      }
      if (tep->_grade > maxgrade) { tep->_grade = maxgrade; }
      if (tep->_grade < mingrade) { tep->_grade = mingrade; }
      _exactsum += tep->_exactgrade;
      _exactsquaredsum += tep->_exactgrade * tep->_exactgrade;
      _sum += tep->_grade;
      _squaredsum += tep->_grade * tep->_grade;
      tp = tp->_next;
    }
  }
}



void exam::add_min( exam* from, float rounding, float maxgrade, float mingrade)
{ //       =======
  if (rounding == -1) {
    //we are in the process of finding the minimum grade
    if (_cglp) {
      _cglp->put_entry(from);
    } else {
      _cglp = new computegradelist( computegradelist::ismin );
      _cglp->put_entry(from);
    }
    simpleListEntry* fp = from->_first;
    exam_entry* fep;
    exam_entry* tep;
    while (fp) {
      fep = (exam_entry*)fp;
      tep = stud_in_exam_search((fep->_studp)->strngp);
      if ( tep != 0 ) {
	if ( tep->_exactgrade > fep->_grade ) {
	  tep->_exactgrade = fep->_grade;
	}
      } else {
	exam_entry* eep = new exam_entry;
	if ( eep==0 ) memory_error();
	append(eep);
	eep->_studp = fep->_studp;
	eep->_examp = this;
	eep->_exactgrade = fep->_grade;
	student_entry* sep = new student_entry;
	if ( sep==0 ) memory_error();
	sep->eep = eep;
	(fep->_studp)->append(sep);
      }
      fp = fp->_next;
    }
  } else {
    //finding the maximum has ended, now we need to round if required
    simpleListEntry* tp = _first;
    exam_entry* tep;
    while (tp) {
      tep = (exam_entry*)tp;
      if (rounding>0) {
	tep->_grade = round((tep->_exactgrade)/rounding)*rounding;
      } else {
	tep->_grade = tep->_exactgrade;
      }
      if (tep->_grade > maxgrade) { tep->_grade = maxgrade; }
      if (tep->_grade < mingrade) { tep->_grade = mingrade; }
      _exactsum += tep->_exactgrade;
      _exactsquaredsum += tep->_exactgrade * tep->_exactgrade;
      _sum += tep->_grade;
      _squaredsum += tep->_grade * tep->_grade;
      tp = tp->_next;
    }
  }
}



void exam::clear()
{ //       =====
  exam_entry *p = (exam_entry*)_first;
  while (p) {
    exam_entry *next = (exam_entry*)p->_next;
    if (p->_examp->isexam()) {
      p->_pointsp->clear();
    }
    delete p;
    p = next;
  }
  _list_length = 0;
  _first = _last = 0;
  _numberofproblems = 0;
  _exactsum = 0;
  _exactsquaredsum = 0;
  _sum = 0;
  _squaredsum = 0;
  if (_pointssump) {
    _pointssump->clear();
  }
  _pointssump = 0;
  if (_cglp) {
    _cglp->clear();
  }
  _cglp = 0;
}



void exam::clear_comment()
{ //       =============
    if ( _comment!=empty_string ) delete[] _comment;
}



void exam::print( ostream& os, int extraspace )
{ //       =====
  const char *margin = "";
  if (extraspace & 1) { 
    // margin = "    ";   // as you please
    os << endl << endl;
  }
  os << margin << strngp << ": " << _comment << endl;
  os << margin << "weight: " << _weight << endl;	
  if (extraspace & 1) { os << endl; }
  if ( _isexam == 1 ) {
    // exam
    os << margin << "points->grade: " << adjust_right
       << setw(4) << _x1 << setw(4) << _y1 << endl;
    os << margin << "               " << adjust_right
       << setw(4) << _x2 << setw(4) << _y2 << endl;
    if (extraspace & 1) { os << endl; }
    simpleListEntry* p = _first;
    simpleListEntry* pep;
    exam_entry* eep;
    while (p) {
      eep = (exam_entry*) p;
      os << margin << adjust_left
	 << setw(10) << eep->_studp->strngp 
	 << adjust_right;
      pep = eep->_pointsp->_first;
      int NoP = 0;
      while (pep) {
	os << setw(7) << ((points_entry*)pep)->_points;
	NoP += 1;
	pep=pep->_next;
      }
      while ( NoP < _numberofproblems ) {
	os << setw(7) << "   ";
	NoP += 1;
      }
      os << setw(10) << eep->_pointsp->totalpoints
	 << adjust_left << "    " << setw(10) << eep->_exactgrade
	 << setw(10) << eep->_grade << endl;
      p = p->_next;
    }
    os << endl << margin << adjust_left
       << setw(10) << "average:" << adjust_right;
    pep = _pointssump->_first;
    while (pep) {
      os << setw(7)
	 << round(100*((points_entry*)pep)->_points/listLength())/100;
      pep=pep->_next;
    }
    os << setw(10) << round(_pointssump->totalpoints/listLength()*100)/100
       << adjust_left << "    " 
       << setw(10) << _exactsum/listLength() 
       << setw(10) << _sum/listLength() << endl;
    os << margin << adjust_left
       << setw(10) << "stdev: " << adjust_right;
    for (int NoP = 0; NoP < _numberofproblems; NoP++) {
      os << setw(7) << "   ";
    }
    os << setw(10) << "   " << adjust_left << "    " << setw(10)
       << sqrt(_exactsquaredsum/listLength() - _exactsum*_exactsum/(listLength()*listLength()))
       << setw(10)
       << sqrt(_squaredsum/listLength() - _sum*_sum/(listLength()*listLength()))
       << endl;
  } else {
    // grade
    if (_cglp) {
      if ( _cglp->_type == computegradelist::isavg ) {
	os << margin << adjust_left
	   << setw(10) << "AVG of:";
      } else if ( _cglp->_type == computegradelist::ismax ) {
	os << margin << adjust_left
	   << setw(10) << "MAX of:";
      } else {
	os << margin << adjust_left
	   << setw(10) << "MIN of:";
      }
      simpleListEntry* glp = _cglp->_first;
      computegrade_entry* gep;
      while (glp) {
	gep = (computegrade_entry*) glp;
	os << setw(8) << gep->_gradep->strngp;
	glp = glp->_next;
      }
      os << " >   " << setw(10) << "exact"
	 << setw(8) << "grade" << endl;
      if ( _cglp->_type == computegradelist::isavg ) {
	os <<  margin << adjust_left
	   << setw(10) << " ";
	glp = _cglp->_first;
	while (glp) {
	  gep = (computegrade_entry*) glp;
	  os << setw(8) << gep->_gradep->weight();
	  glp = glp->_next;
	}
	os << endl;
      } else {
	os << endl;
      }
    }
    simpleListEntry* p = _first;
    exam_entry* eep;
    while (p) {
      eep = (exam_entry*) p;
      os << margin << adjust_left
	 << setw(10) << eep->_studp->strngp;
      if (_cglp) {
	simpleListEntry* glp = _cglp->_first;
	computegrade_entry* gep;
	exam_entry* cgleep;
	while (glp) {
	  gep = (computegrade_entry*) glp;
	  cgleep = gep->_gradep->stud_in_exam_search(eep->_studp->strngp);
	  if (cgleep) {
	    os << setw(8) << cgleep->_grade;
	  } else {
	    os << setw(8) << " ";
	  }
	  glp = glp->_next;
	}
	os << " >   ";
      }
      os << setw(10) << eep->_exactgrade 
	 << setw(8) << eep->_grade << endl;
      p = p->_next;
    }
    os << endl << margin << adjust_left
       << setw(10) << "average: " << adjust_left;
    if (_cglp) {
      simpleListEntry* glp = _cglp->_first;
      while (glp) {
	os << setw(8) << " ";
	glp = glp->_next;
      }
      os << "     ";
    }
    os << setw(10) << _exactsum/listLength()
       << setw(10) << _sum/listLength() << endl;
    os << margin << adjust_left
       << setw(10) << "stdev: ";
    if (_cglp) {
      simpleListEntry* glp = _cglp->_first;
      while (glp) {
	os << setw(8) << " ";
	glp = glp->_next;
      }
      os << "     ";
    }
    os << setw(10)
       << sqrt(_exactsquaredsum/listLength() - _exactsum*_exactsum/(listLength()*listLength()))
     << setw(8)
       << sqrt(_squaredsum/listLength() - _sum*_sum/(listLength()*listLength()))
       << endl;
  }
}



void exam::print_LaTeX( ostream& os, int naming, int landscape,
/*         ===========*/ char* classname ) 
{
  // naming: 0=name, 1=comment, 2=anonymous (no name, no exact grades)
  int columns;
  if ( landscape==0 ) {
    os << latex_header1 << latex_header2;
  } else {
    os << latex_header1l << latex_header2;
  }
  if (isexam()) {
    // EXAM*****************************************************
    if ( naming < 2 ) {
      columns = _numberofproblems+4;
      os << "\\begin{center}\n  \\begin{tabular}{@{}l*{"
	 << _numberofproblems << "}{r}rll@{}}\n    ";
      if (classname[0]) {
	os << "\\multicolumn{" << columns
	   << "}{@{}c@{}}{\\Large\\textbf{" << classname
	   << "}}\\\\[3mm]\n    ";
      }
      os << "\\toprule\n    \\multicolumn{" << columns
	 << "}{@{}l@{}}{\\large\\textbf{";
      os << strngp << ": \\;" << _comment << "}}\\\\\n    "
	 << "\\textbf{Name} & ";
    } else {
      columns = _numberofproblems+3;
      os << "\\begin{center}\n  \\begin{tabular}{@{}l*{"
	 << _numberofproblems << "}{r}rl@{}}\n    ";
      if (classname[0]) {
	os << "\\multicolumn{" << columns
	   << "}{@{}c@{}}{\\Large\\textbf{" << classname
	   << "}}\\\\[3mm]\n    ";
      }
      os << "\\toprule\n    \\multicolumn{" << columns
	 << "}{@{}l@{}}{\\large\\textbf{";
      os << strngp << ": \\;" << _comment << "}}\\\\\n    "
	 << "& ";
    }
    for (int i=1; i<=_numberofproblems; i++) {
      os << "\\textbf{" << i << "} & ";
    }
    if ( naming < 2 ) {
      os << "\n    \\textbf{Total} & \\textbf{Auto} & "
	 << "\\textbf{Grade} \\\\\\midrule\n    ";
    } else {
      os << "\n    \\textbf{Total} & "
	 << "\\textbf{Grade} \\\\\\midrule\n    ";
    }
    simpleListEntry* p = _first;
    simpleListEntry* pep;
    exam_entry* eep;
    while (p) {
      eep = (exam_entry*) p;
      if (naming==0) {
	os << eep->_studp->strngp;
      } else if (naming==1) { 
	os << eep->_studp->comment();
      }
      os << " & ";
      pep = eep->_pointsp->_first;
      int NoP = 0;
      while (pep) {
	os << ((points_entry*)pep)->_points << " & ";
	NoP += 1;
	pep=pep->_next;
      }
      while ( NoP < _numberofproblems ) {
	os << " & ";
	NoP += 1;
      }
      os << "\\textbf{" << eep->_pointsp->totalpoints
	 << "} & ";
      if ( naming < 2 ) {
	os << eep->_exactgrade << " & ";
      }
      os << "\\textbf{" << eep->_grade << "} \\\\\n    "; 
      p = p->_next;
    }
    os << "\\midrule\n    ";
    os << "Average & ";
    pep = _pointssump->_first;
    while (pep) {
      os << round(100*((points_entry*)pep)->_points/listLength())/100
	 << " & ";
      pep=pep->_next;
    }
    os << "\\textbf{"
       << round(_pointssump->totalpoints/listLength()*100)/100
       << "} & ";
    if ( naming < 2 ) {
      os << _exactsum/listLength() << " & ";
    } 
    os << "\\textbf{" << _sum/listLength() << "}\\\\\n    ";
    if ( naming < 2 ) {
      os << "\\multicolumn{" << columns-2 << "}{@{}l@{}}{StDev} & "
	 << sqrt(_exactsquaredsum/listLength() - _exactsum*_exactsum/(listLength()*listLength()))
	 << " & ";
    } else {
      os << "\\multicolumn{" << columns-1 << "}{@{}l@{}}{StDev} & ";
    }
    os << sqrt(_squaredsum/listLength() - _sum*_sum/(listLength()*listLength()))
       << "\\\\\n    ";
    os << "\\multicolumn{" << columns << "}{@{}l@{}}{Weight: \\;"
       << _weight << "}\\\\\n    ";
    os << "\\multicolumn{" << columns
       << "}{@{}l@{}}{points $\\to$ grade: \\;"
       << _x1 << " \\quad " << _y1 << "}\\\\\n    ";
    os << "\\multicolumn{" << columns
       << "}{@{}l@{}}{\\phantom{points $\\to$ grade: \\;}"
       << _x2 << " \\quad " << _y2 << "}\\\\\n    ";
    os << "\\bottomrule\n  \\end{tabular}\n\\end{center}\n\n";
  } else {
    // GRADE****************************************************
    columns = 3;
    if (_cglp) {
      columns = _cglp->listLength()+3;
      os << "\\begin{center}\n  \\begin{tabular}{@{}l*{"
	 << columns-3 << "}{l}ll@{}}\n    ";
      if (classname[0]) {
	os << "\\multicolumn{" << columns
	   << "}{@{}c@{}}{\\Large\\textbf{" << classname
	   << "}}\\\\[3mm]\n    ";
      }
      os << "\\toprule\n    \\multicolumn{" << columns
	 << "}{@{}l@{}}{\\large\\textbf{"
	 << strngp << ": \\;" << _comment << "}}\n    \\\\\n    ";
      if ( _cglp->_type == computegradelist::isavg ) {
	os << "average of: & ";
      } else if ( _cglp->_type == computegradelist::ismax ) {
	os << "maximum of: & ";
      } else {
	os << "minimum of: & ";
      }
      simpleListEntry* glp = _cglp->_first;
      computegrade_entry* gep;
      while (glp) {
	gep = (computegrade_entry*) glp;
	os << gep->_gradep->strngp << " & ";
	glp = glp->_next;
      }
      os << "\\textbf{Exact} & \\textbf{Grade} \\\\\n    ";
      if ( _cglp->_type == computegradelist::isavg ) {
	os << "& ";
	glp = _cglp->_first;
	while (glp) {
	  gep = (computegrade_entry*) glp;
	  os << "(" << gep->_gradep->weight() << ") & ";
	  glp = glp->_next;
	}
	os << " & \\\\\n    ";
      }
      os << "\\midrule\n    ";
    } else {
      os << "\\begin{center}\n  \\begin{tabular}{@{}lll@{}}\n    ";
      if (classname[0]) {
	os << "\\multicolumn{3}{@{}c@{}}{\\Large\\textbf{" << classname
	   << "}}\\\\[3mm]\n    ";
      }
      os << "\\toprule\n    \\multicolumn{3}{@{}l@{}}{\\large\\textbf{"
	 << strngp << ": \\;" << _comment << "}}\n    \\\\\n    ";
      if ( naming < 2 ) {
	os << "\\textbf{Name} ";
      }
      os << "& \\textbf{Exact} & \\textbf{Grade} \\\\\\midrule\n    ";
    }
    simpleListEntry* p = _first;
    exam_entry* eep;
    while (p) {
      eep = (exam_entry*) p;
      if (naming==0) {
	os << eep->_studp->strngp << " ";
      } else if (naming==1) { 
	os << eep->_studp->comment() << " ";
      }
      if (_cglp) {
	simpleListEntry* glp = _cglp->_first;
	computegrade_entry* gep;
	exam_entry* cgleep;
	while (glp) {
	  gep = (computegrade_entry*) glp;
	  cgleep = gep->_gradep->stud_in_exam_search(eep->_studp->strngp);
	  if (cgleep) {
	    os << " & " << cgleep->_grade;
	  } else {
	    os << " & ";
	  }
	  glp = glp->_next;
	}
      }
      os << " & \\textbf{" << eep->_exactgrade << "} & \\textbf{"
	 << eep->_grade << "} \\\\\n    "; 
      p = p->_next;
    }
    os << "\\midrule\n    ";
    os << "\\multicolumn{" << columns-2 << "}{@{}l}{Average} & "
       << _exactsum/listLength() << " & "
       << _sum/listLength() << " \\\\\n    ";
    os << "\\multicolumn{" << columns-2 << "}{@{}l}{StDev} & ";
    os << sqrt(_exactsquaredsum/listLength() - _exactsum*_exactsum/(listLength()*listLength()))
       << " & "
       << sqrt(_squaredsum/listLength() - _sum*_sum/(listLength()*listLength()))
       << " \\\\\n    ";
    os << "\\multicolumn{" << columns-1 << "}{@{}l}{Weight} & "
       << _weight << " \\\\\n    ";
    os << "\\bottomrule\n  \\end{tabular}\n\\end{center}\n\n";
  }
  os << latex_footer;
}



void exam_list::print_exam_list( ostream& os, int tab, int width )
{ //            ================
  simpleListEntry* p = _first;
  exam* ep;
  os << adjust_left; 
  while (p) {
    ep = (exam*) p;
    os << setw(15) << ep->strngp;
    if ( ep->isexam() ) {
      os << setw(18) << "exam with weight";
    } else {
      os << setw(18) << "grade with weight";
    }    
    os << setw(10) << ep->weight()
       << ep->comment() << endl;
    p = p->_next;
  }
}



void exam_list::print_exams_for_student( ostream& os, const char* st,
/*              ======================= */ int width )
{
  simpleListEntry* p = _first;
  exam* ep;
  exam_entry* eep;
  while (p) {
    ep = (exam*) p;
    if (ep->ishidden() == 0) {
      if ( st == 0 ) {
	os << setw(width) << ep->strngp;
      } else {
	eep = ep->stud_in_exam_search(st);
	if (eep) {
	  os << setw(width) << eep->_grade;
	} else {
	  os << setw(width) << "";
	}
      }
    }
    p = p->_next;
  }
}



int exam_list::print_exams_for_student_latex(  ostream& os, const char* st,
/*             =============================*/ int mode )
{ // mode=0: return number of nonempty exams/grades
  // mode=1: print header line of table
  // mode=2: print body of table 
  // mode=3: print averages 
  // mode=4: print weights 
  simpleListEntry* p = _first;
  exam* ep;
  exam_entry* eep;
  int counter=0;
  while (p) {
    ep = (exam*) p;
    if (ep->ishidden() == 0) {
      if ( mode == 0 ) {
	counter++;
      } else if ( mode == 1 ) {
	os << " & \\textbf{" << ep->strngp << "}";
      } else if ( mode == 2 ) {
	eep = ep->stud_in_exam_search(st);
	if (eep) {
	  os << " & " << eep->_grade;
	} else {
	  os << " &";
	}
      } else if ( mode == 3 ) {
	os << " & " << round(ep->average()*100)/100;
      }  else if ( mode == 4 ) {
	os << " & " << ep->weight();
      }
    }
    p = p->_next;
  }
  return counter;
}



void exam_list::print_latex_tofiles( const char* folder, int naming, 
/*              ===================*/ int landscape, char* classname )
{ // naming: 0=name, 1=comment, 2=anonymous (no names, no exact grades)
  string filenamesuffix = ".tex";
  simpleListEntry* p = _first;
  exam* ep;
  while (p) {
    ep = (exam*) p;
    ofstream file(string(folder) + "/" + ep->strngp + filenamesuffix);
    if (file.is_open()) {
      ep->print_LaTeX(file,naming,landscape,classname);
      file.close();
    } else { 
      cerr << "failure" << endl;
      return;
    }
    p = p->_next;
  }
}



void exam_list::clear( int delete_account_names )
{ //            =====
    exam *p = (exam*)_first;
    while (p) {
        exam *next = (exam*)p->_next;
	p->clear();
        if (delete_account_names) delete[] (char*)p->strngp;
	delete p;
        p = next;
    }
    _list_length = 0;
    _first = _last = 0;
}



student::student( char* name, char* comment )
{ //     =======
  strngp = name; 
  if ( comment==0 ) {
    _comment = empty_string;
    _indexcomment = empty_string;
  } else {
    _comment = new char[strlen(comment)+1];
    if ( _comment==0 ) memory_error();
    strcpy( _comment, comment );
    _indexcomment = new char[strlen(comment)+1];
    if ( _indexcomment==0 ) memory_error();
    const int numspec=14;
    const char* umlaut[numspec] = {"ü", "Ü", "ä", "Ä", "ö", "Ö", "ë", "Ë", "é", "É", "è", "È", "ê", "Ê"};
    const char* replace = "uuaaooeeeeeeee";
    int innerflag;
    int outerflag;
    int loop;
    int i=0;
    int k=0;
    while (comment[i]) {
      if ( (comment[i] >= 'a') && (comment[i] <= 'z') ) {
	_indexcomment[k]=comment[i];
      } else if ( (comment[i] >= 'A') && (comment[i] <= 'Z') ) {
	_indexcomment[k]=comment[i]-'A'+'a';
      } else {
	outerflag=0;
	loop=0;
	while ( loop < numspec) {
	  innerflag=1;
	  for (int j=0; j<strlen(umlaut[loop]); j++) {
	    if (comment[i+j] != umlaut[loop][j]) {
	      innerflag=0;
	      break;
	    }
	  }
	  if (innerflag == 1) {
	    _indexcomment[k]=replace[loop];
	    i=i+strlen(umlaut[loop])-1;
	    outerflag=1;
	    break;
	  }
	  loop++;
	}
	if (outerflag == 0 ) {
	  _indexcomment[k]=comment[i];
	}
      }
      i++;
      k++;
    }
  }
}



void student::clear()
{ //          =====
    student_entry *p = (student_entry*)_first;
    while (p) {
        student_entry *next = (student_entry*)p->_next;
        delete p;
        p = next;
    }
    _list_length = 0;
    _first = _last = 0;
}



void student::clear_comment()
{ //          =============
    if ( _comment!=empty_string ) delete[] _comment;
    if ( _indexcomment!=empty_string ) delete[] _indexcomment;
}



void student::print( ostream& os, int extraspace)
{ //          =====
  const char *margin = "";
  if (extraspace & 1) { 
    // margin = "    ";   // as you please
    os << endl << endl;
  }
  os << margin << strngp << ": " << _comment << endl;
  //os << margin << strngp << ": " << _indexcomment << endl;
  if (extraspace & 1) { os << endl; }
  simpleListEntry* p = _first;
  student_entry* sep;
  exam_entry* eep;
  while (p) {
    sep = (student_entry*) p;
    eep = sep->eep;
    os << margin << adjust_left
       << setw(10) << eep->_examp->strngp 
       << setw(10) << eep->_exactgrade
       << setw(8) << eep->_grade
       << setw(8) << eep->_examp->weight()
       << eep->_examp->comment() << endl;
    p = p->_next;
  }
}



void student::print_LaTeX( ostream& os, int landscape, char* classname ) 
{ //          ===========  
  if ( landscape==0 ) {
    os << latex_header1 << latex_header2;
  } else {
    os << latex_header1l << latex_header2;
  }
  os << "\\begin{center}\n  \\begin{tabular}{@{}lllll@{}}\n    ";
  if (classname[0]) {
    os << "\\multicolumn{5}{@{}c@{}}{\\Large\\textbf{" << classname
       << "}}\\\\[3mm]\n    ";
  }
  os << "\\toprule\n    "
     << "\\multicolumn{5}{@{}l@{}}{\\large\\textbf{"
     << strngp << ": \\;" << _comment << "}}\n    \\\\\\midrule\n    ";
  simpleListEntry* p = _first;
  exam_entry* eep;
  student_entry* sep;
  while (p) {
    sep = (student_entry*) p;
    eep = sep->eep;
    os << eep->_examp->strngp << " & " 
       << eep->_exactgrade << " & "
       << eep->_grade << " & " 
       << eep->_examp->weight() << " & "
       << eep->_examp->comment() << " \\\\\n    ";
    p = p->_next;
  }
  os << "\\bottomrule\n  \\end{tabular}\n\\end{center}\n\n";
  os << latex_footer;
}



void student_list::print_overview( ostream& os, exam_list* elp, int width)
{ //               ==============
  os << adjust_left << setw(12) << " ";
  elp->print_exams_for_student(os, 0, width);
  os << endl;
  simpleListEntry* p = _first;
  student* sp;
  while (p) {
    sp = (student*) p;
    os << setw(12) << sp->strngp;
    elp->print_exams_for_student(os, sp->strngp, width);
    os << endl;
    p = p->_next;
  }
}


void student_list::print_student_list( ostream& os, int tab, int width )
{ //               ==================
  simpleListEntry* p = _first;
  student* sp;
  os << adjust_left; 
  while (p) {
    sp = (student*) p;
    os << setw(15) << sp->strngp << sp->comment() << endl;
    p = p->_next;
  }
}



void student_list::clear( int delete_account_names )
{ //               =====
  student *p = (student*)_first;
  while (p) {
    student *next = (student*)p->_next;
    if (delete_account_names) delete[] (char*)p->strngp;
    delete p;
    p = next;
  }
  _list_length = 0;
  _first = _last = 0;
}



void student_list::print_latex_tofiles(  const char* folder, int naming, 
/*                 ===================*/ int landscape, char* classname,
					 int no_of_fields, exam_list* elp)
{ // naming: 0=name, 1=comment, 2=anonymous
  // files per student
  string filenamesuffix = ".tex";
  simpleListEntry* p = _first;
  student* sp;
  while (p) {
    sp = (student*) p;
    ofstream file(string(folder) + "/" + sp->strngp + filenamesuffix);
    if (file.is_open()) {
      sp->print_LaTeX(file,landscape,classname);
      file.close();
    } else { 
      cerr << "failure" << endl;
      return;
    }
    p = p->_next;
  }
  // file z_studlist with empty fields
  ofstream file(string(folder) + "/z_studlist" + filenamesuffix);
  if (file.is_open()) {
    if ( landscape==0 ) {
      file << latex_header1 << latex_header2;
    } else {
      file << latex_header1l << latex_header2;
    }
    file << "\\begin{center}\n  \\renewcommand{\\arraystretch}{1.4}\n  "
	 << "\\begin{tabular}{@{}|l*{" << no_of_fields
	 << "}{|p{13mm}}|@{}}\n    ";
    if (classname[0]) {
      file << "\\multicolumn{" << no_of_fields + 1
	   << "}{@{}c@{}}{\\Large\\textbf{" << classname
	   << "}}\\\\[5mm]\n    ";
    }
    file << "\\hline\n    \\textbf{Name}";
    for (int i = no_of_fields; i > 0; i--) {
      file << " & \\hfill\\textbf{}\\hfill\\mbox{}";
    }
    file << "\n    \\\\\\hline\\hline\n";
    p = _first;
    while (p) {
      sp = (student*) p;
      if (naming == 1) {
	file << "    " << sp->comment() << " ";
      } else {
	file << "    " << sp->strngp << " ";
      }
      for (int i = no_of_fields; i > 0; i--) {
	file << "&";
      }
      file << " \\\\\\hline\n";
      p = p->_next;
    }
    file <<  "  \\end{tabular}\n\\end{center}\n\n"
	  << latex_footer;      
    file.close();
  } else { 
    cerr << "failure" << endl;
    return;
  }
  // file z_overview
  ofstream newfile(string(folder) + "/z_overview" + filenamesuffix);
  if (newfile.is_open()) {
    if ( landscape==0 ) {
      newfile << latex_header1 << latex_header2;
    } else {
      newfile << latex_header1l << latex_header2;
    }
    int columns = elp->print_exams_for_student_latex(newfile, 0, 0);
    newfile << "\\begin{center}\n  \\begin{tabular}{@{}l*{" << columns
	    << "}{l}@{}}\n    ";
    if (classname[0]) {
      newfile << "\\multicolumn{" << columns + 1
	      << "}{@{}c@{}}{\\Large\\textbf{" << classname
	      << "}}\\\\[3mm]\n    ";
    }
    newfile << "\\toprule\n    \\multicolumn{" << columns+1
	    << "}{@{}l@{}}{\\large\\textbf{Overview}}\\\\\n    ";
    //<< "\\textbf{Name}";
    elp->print_exams_for_student_latex(newfile, 0, 1);
    newfile << " \\\\\n    \\midrule\n    ";
    p = _first;
    int studcounter=0;
    while (p) {
      sp = (student*) p;
      if ( columns > 9 ) {
	studcounter++;
      }
      if (naming == 1) {
	newfile << sp->comment();
      } else if (naming == 0) { 
	newfile << sp->strngp;
      }
      elp->print_exams_for_student_latex(newfile, sp->strngp, 2);
      if (studcounter >= 4) {
	studcounter=0;
	newfile << " \\\\\\midrule\n    ";
      } else {
	newfile << " \\\\\n    ";
      }
      p = p->_next;
    }
    if ( columns > 9 ) {
      newfile << "\\midrule";
    }
    newfile << "\\midrule\n    Average";
    elp->print_exams_for_student_latex(newfile, 0, 3);
    newfile << " \\\\\n    Weight";
    elp->print_exams_for_student_latex(newfile, 0, 4);
    newfile << " \\\\\n    \\bottomrule\n  \\end{tabular}\n\\end{center}\n\n"
	 << latex_footer;      
    newfile.close();
  } else { 
    cerr << "failure" << endl;
    return;
  }
}


void student_list::sort( int order )
{ //               ====
  simpleListEntry* oldcurrentp = _first;
  simpleListEntry* oldnextp;
  _first = 0;
  _last = 0;
  _list_length = 0;
  simpleListEntry* currentp;
  simpleListEntry* previousp;
  student* sp;
  student* oldsp;
  while (oldcurrentp) {
    oldnextp = oldcurrentp->_next;
    oldsp = (student*)oldcurrentp;
    currentp = _first;
    previousp = 0;
    while (currentp) {
      sp = (student*)currentp;
      if ( order == exam::sortbyname ) {
	if (stringcomp(oldsp->strngp,sp->strngp)) {
	  break;
	}
      } else if ( order == exam::sortbycomment ) {
	if (stringcomp(oldsp->index(),sp->index())) {
	  break;
	}
      } else {
	break;
      }
      previousp = currentp;
      currentp = currentp->_next;
    }
    insert(oldcurrentp,previousp);
    oldcurrentp = oldnextp;
  }
}


int student_list::stringcomp( const char* first, const char* second )
{ //              ==========
  // returns 1 if first <= second, 0 otherwise
  int i=0;
  char firsti, secondi;
  while ( (firsti=first[i]) == (secondi=second[i]) ) {
    if (firsti) i++; 
    else return 1; 
  }
  if ( firsti < secondi ) return 1;
  return 0;
}
