// lineInput.cc          Lg, July 19, 1995
// version 2.0           SM, May 2021

#include "lineInput.h"
#include "uti1a.h"

int li_get( istream& is, char& c )
{ //======
    char ch;
    if ( !is.get(ch) ) return 0;
    if ( (ch=='\n') || (ch=='\r') ) {
        is.putback(ch);
        return 0;
    }
    c = ch;
    return 1;
}


int li_eatspace( istream& is )
{ //===========
    char ch;
    int at_least_one_space = 0;
    for (;;) {
        if ( !li_get(is,ch) ) break;
        if ( (ch!=' ') && (ch!='\t') ) { is.putback(ch); break; }
        at_least_one_space = 1;
    }
    return at_least_one_space;
}


static int li_remove_trailing_space( char* s )
{ //       ------------------------
    int i=0;
    int j=0;
    char ch;
    while ( (ch=s[j++])!=0 )
        if ( ( ch!=' ') && (ch!='\t') ) i=j;
    s[i]=0;
    return i;
}


int li_get_string( istream& is, char* buf, int bufsize, 
/*  ============= */            int (*is_allowed)(char), 
			        int line_count,      // for error message
                                int remove_trailing_space )
{
    int nchar=0;
    char ch;
    for (;;) {
        if ( !li_get(is,ch) ) break;
        if ( !is_allowed(ch) ) { is.putback(ch); break; }
        buf[nchar++]=ch;
        if ( nchar >= bufsize ) { 
            input_error( line_count, "input string buffer overflow" );
            is.putback(ch);
            buf[bufsize-1]=0;
            return bufsize; 
        }
    }
    buf[nchar]=0;
    if ( remove_trailing_space ) return li_remove_trailing_space(buf);
    else return nchar;
}

int li_get_float( istream& is, float& x )
{ //============
  float y=0;
  int flag=0; // 0=nothing read, 1=sign read,
  // 2=digits read, 3=decimal point read
  char ch;
  float digits=0;
  int sign=1;
  li_eatspace(is);
  for (;;) {
    if ( !li_get(is,ch) ) break;
    if ( !isdigit(ch) ) {
      if ( (ch == '.') && (flag<3) ) {
	digits=1;
	flag=3;
      } else if ( (ch=='+') && (flag<1) ) {
	flag=1;
      } else if ( (ch=='-') && (flag<1) ) {
	flag=1;
	sign=-1;
      } else if ( (ch!=' ') && (ch!='\t') ) {
	flag=0; break;
      } else {
	break;
      }
    } else {
      if ( flag<2 ) {
	flag=2;
      }
      y = 10*y + (ch-'0');
      digits = digits * 10;
    }
  }
  if ( flag==2 ) {
    x=y*sign;
    return 1;
  } else if (flag==3) {
    x=y/digits*sign;
    return 2;
  }
  return 0;
}

